import plotly.colors
import math

class Color:
    @staticmethod
    def ByValueInRange(value: float = 0.5, minValue: float = 0.0, maxValue: float = 1.0, alpha: float = 1.0, useAlpha: float = False, colorScale="viridis"):
        """
        Retursn the r, g, b, (and optionally) a list of numbers representing the red, green, blue and alpha color elements.
        Parameters
        ----------
        value : float , optional
            The input value. The default is 0.5.
        minValue : float , optional
            the input minimum value. The default is 0.0.
        maxValue : float , optional
            The input maximum value. The default is 1.0.
        alpha : float , optional
            The int alpha (transparency) value. 0.0 means the color is fully transparent, 1.0 means the color is fully opaque. The default is 1.0.
        useAlpha : bool , optional
            If set to True, the returns list includes the alpha value as a fourth element in the lists.
        colorScale : str , optional
            The desired type of plotly color scales to use (e.g. "Viridis", "Plasma"). The default is "Viridis". For a full list of names, see https://plotly.com/python/builtin-colorscales/.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        
        # Code based on: https://stackoverflow.com/questions/62710057/access-color-from-plotly-color-scale

        def hex_to_rgb(value):
            value = str(value)
            value = value.lstrip('#')
            lv = len(value)
            returnValue = tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))
            return str(returnValue)

        def get_color(colorscale_name, loc):
            from _plotly_utils.basevalidators import ColorscaleValidator
            # first parameter: Name of the property being validated
            # second parameter: a string, doesn't really matter in our use case
            cv = ColorscaleValidator("colorscale", "")
            # colorscale will be a list of lists: [[loc1, "rgb1"], [loc2, "rgb2"], ...] 
            colorscale = cv.validate_coerce(colorscale_name)
            
            if hasattr(loc, "__iter__"):
                return [get_continuous_color(colorscale, x) for x in loc]
            color = get_continuous_color(colorscale, loc)
            color = color.replace("rgb", "")
            color = color.replace("(", "")
            color = color.replace(")", "")
            color = color.split(",")
            final_colors = []
            for c in color:
                final_colors.append(math.floor(float(c)))
            return final_colors

        def get_continuous_color(colorscale, intermed):
            """
            Plotly continuous colorscales assign colors to the range [0, 1]. This function computes the intermediate
            color for any value in that range.

            Plotly doesn't make the colorscales directly accessible in a common format.
            Some are ready to use:
            
                colorscale = plotly.colors.PLOTLY_SCALES["Greens"]

            Others are just swatches that need to be constructed into a colorscale:

                viridis_colors, scale = plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.Viridis)
                colorscale = plotly.colors.make_colorscale(viridis_colors, scale=scale)

            :param colorscale: A plotly continuous colorscale defined with RGB string colors.
            :param intermed: value in the range [0, 1]
            :return: color in rgb string format
            :rtype: str
            """
            if len(colorscale) < 1:
                raise ValueError("colorscale must have at least one color")
            if intermed <= 0 or len(colorscale) == 1:
                c = colorscale[0][1]
                return c if c[0] != "#" else hex_to_rgb(c)
            if intermed >= 1:
                c = colorscale[-1][1]
                return c if c[0] != "#" else hex_to_rgb(c)
            for cutoff, color in colorscale:
                if intermed > cutoff:
                    low_cutoff, low_color = cutoff, color
                else:
                    high_cutoff, high_color = cutoff, color
                    break
            if (low_color[0] == "#") or (high_color[0] == "#"):
                # some color scale names (such as cividis) returns:
                # [[loc1, "hex1"], [loc2, "hex2"], ...]
                low_color = hex_to_rgb(low_color)
                high_color = hex_to_rgb(high_color)
            return plotly.colors.find_intermediate_color(
                lowcolor=low_color,
                highcolor=high_color,
                intermed=((intermed - low_cutoff) / (high_cutoff - low_cutoff)),
                colortype="rgb",
            )
        
        def get_color_default(ratio):
            r = 0.0
            g = 0.0
            b = 0.0

            finalRatio = ratio;
            if (finalRatio < 0.0):
                finalRatio = 0.0
            elif(finalRatio > 1.0):
                finalRatio = 1.0

            if (finalRatio >= 0.0 and finalRatio <= 0.25):
                r = 0.0
                g = 4.0 * finalRatio
                b = 1.0
            elif (finalRatio > 0.25 and finalRatio <= 0.5):
                r = 0.0
                g = 1.0
                b = 1.0 - 4.0 * (finalRatio - 0.25)
            elif (finalRatio > 0.5 and finalRatio <= 0.75):
                r = 4.0*(finalRatio - 0.5);
                g = 1.0
                b = 0.0
            else:
                r = 1.0
                g = 1.0 - 4.0 * (finalRatio - 0.75)
                b = 0.0

            rcom =  (max(min(r, 1.0), 0.0))
            gcom =  (max(min(g, 1.0), 0.0))
            bcom =  (max(min(b, 1.0), 0.0))

            return [rcom,gcom,bcom]
        
        if minValue > maxValue:
            temp = minValue;
            maxValue = minValue
            maxValue = temp

        val = value
        val = max(min(val,maxValue), minValue) # bracket value to the min and max values
        if (maxValue - minValue) != 0:
            val = (val - minValue)/(maxValue - minValue)
        else:
            val = 0
        if not colorScale or colorScale.lower() == "default":
            rgbList = get_color_default(val)
        else:
            rgbList = get_color(colorScale, val)
        if useAlpha:
            rgbList.append(alpha)
        return rgbList



# Vertex Unit testing
import topologicpy
import topologic
from topologicpy.Vertex import Vertex
from topologicpy.CellComplex import CellComplex

v = Vertex.ByCoordinates(10, 20, 30)
assert isinstance(v, topologic.Vertex), "Vertex.ByCoordinates. Should be topologic.Vertex"

v = Vertex.ByCoordinates(10.123, 20.456, 30.789)
x = Vertex.Coordinates(v, outputType="x", mantissa=3)
assert x == [10.123], "Vertex.Coordinates. Should be 10.123"
y = Vertex.Coordinates(v, outputType="y", mantissa=3)
assert y == [20.456], "Vertex.Coordinates. Should be 20.456"
z = Vertex.Coordinates(v, outputType="z", mantissa=3)
assert z == [30.789], "Vertex.Coordinates. Should be 30.789"
xyz = Vertex.Coordinates(v, outputType="xyz", mantissa=3)
assert xyz == [10.123, 20.456, 30.789], "Vertex.Coordinates. Should be 30.789"

v1 = Vertex.ByCoordinates(10, 20, 30)
v2 = Vertex.ByCoordinates(20, 30, 40)
d = Vertex.Distance(v1, v2, mantissa=3)
assert d == 17.321, "Vertex.Distance. Should be 17.321"

origin = Vertex.ByCoordinates(0,0,0)
cc = CellComplex.Prism(origin=origin, width=10, length=10, height=10, uSides=2, vSides=2, wSides=2,
                         direction=[0,0,1], originLocation="Center")
assert isinstance(cc, topologic.CellComplex), "Vertex.EnclosingCell. Should be topologic.CellComplex"
EnclosingCell(vertex, topology, exclusive=True, tolerance=0.0001)


import topologic

def internalVertex(topology, tolerance):
	vst = None
	classType = topology.Type()
	if classType == 64: #CellComplex
		tempCells = []
		_ = topology.Cells(None,tempCells)
		tempCell = tempCells.front()
		vst = topologic.CellUtility.InternalVertex(tempCell, tolerance)
	elif classType == 32: #Cell
		vst = topologic.CellUtility.InternalVertex(topology, tolerance)
	elif classType == 16: #Shell
		tempFaces = []
		_ = topology.Faces(None, tempFaces)
		tempFace = tempFaces.front()
		vst = topologic.FaceUtility.InternalVertex(tempFace, tolerance)
	elif classType == 8: #Face
		vst = topologic.FaceUtility.InternalVertex(topology, tolerance)
	elif classType == 4: #Wire
		if topology.IsClosed():
			internalBoundaries = []
			tempFace = topologic.Face.ByExternalInternalBoundaries(topology, internalBoundaries)
			vst = topologic.FaceUtility.InternalVertex(tempFace, tolerance)
		else:
			tempEdges = []
			_ = topology.Edges(None, tempEdges)
			vst = topologic.EdgeUtility.PointAtParameter(tempVertex.front(), 0.5)
	elif classType == 2: #Edge
		vst = topologic.EdgeUtility.PointAtParameter(topology, 0.5)
	elif classType == 1: #Vertex
		vst = topology
	else:
		vst = topology.CenterOfMass()
	return vst

def listAttributeValues(listAttribute):
	listAttributes = listAttribute.ListValue()
	returnList = []
	for attr in listAttributes:
		if isinstance(attr, topologic.IntAttribute):
			returnList.append(attr.IntValue())
		elif isinstance(attr, topologic.DoubleAttribute):
			returnList.append(attr.DoubleValue())
		elif isinstance(attr, topologic.StringAttribute):
			returnList.append(attr.StringValue())
	return returnList

def getValues(item):
	keys = item.Keys()
	returnList = []
	for key in keys:
		try:
			attr = item.ValueAtKey(key)
		except:
			raise Exception("Dictionary.Values - Error: Could not retrieve a Value at the specified key ("+key+")")
		if isinstance(attr, topologic.IntAttribute):
			returnList.append(attr.IntValue())
		elif isinstance(attr, topologic.DoubleAttribute):
			returnList.append(attr.DoubleValue())
		elif isinstance(attr, topologic.StringAttribute):
			returnList.append(attr.StringValue())
		elif isinstance(attr, topologic.ListAttribute):
			returnList.append(listAttributeValues(attr))
		else:
			returnList.append("")
	return returnList

def getKeys(item):
	stlKeys = item.Keys()
	returnList = []
	copyKeys = stlKeys.__class__(stlKeys) #wlav suggested workaround. Make a copy first
	for x in copyKeys:
		k = x.c_str()
		returnList.append(k)
	return returnList

def getValueAtKey(item, key):
	try:
		attr = item.ValueAtKey(key)
	except:
		raise Exception("Dictionary.ValueAtKey - Error: Could not retrieve a Value at the specified key ("+key+")")
	if isinstance(attr, topologic.IntAttribute):
		return (attr.IntValue())
	elif isinstance(attr, topologic.DoubleAttribute):
		return (attr.DoubleValue())
	elif isinstance(attr, topologic.StringAttribute):
		return (attr.StringValue())
	elif isinstance(attr, topologic.ListAttribute):
		return (listAttributeValues(attr))
	else:
		return None

def processKeysValues(keys, values):
	if len(keys) != len(values):
		raise Exception("DictionaryByKeysValues - Keys and Values do not have the same length")
	stl_keys = []
	stl_values = []
	for i in range(len(keys)):
		if isinstance(keys[i], str):
			stl_keys.append(keys[i])
		else:
			stl_keys.append(str(keys[i]))
		if isinstance(values[i], list) and len(values[i]) == 1:
			value = values[i][0]
		else:
			value = values[i]
		if isinstance(value, bool):
			if value == False:
				stl_values.append(topologic.IntAttribute(0))
			else:
				stl_values.append(topologic.IntAttribute(1))
		elif isinstance(value, int):
			stl_values.append(topologic.IntAttribute(value))
		elif isinstance(value, float):
			stl_values.append(topologic.DoubleAttribute(value))
		elif isinstance(value, str):
			stl_values.append(topologic.StringAttribute(value))
		elif isinstance(value, list):
			l = []
			for v in value:
				if isinstance(v, bool):
					l.append(topologic.IntAttribute(v))
				elif isinstance(v, int):
					l.append(topologic.IntAttribute(v))
				elif isinstance(v, float):
					l.append(topologic.DoubleAttribute(v))
				elif isinstance(v, str):
					l.append(topologic.StringAttribute(v))
			stl_values.append(topologic.ListAttribute(l))
		else:
			raise Exception("Error: Value type is not supported. Supported types are: Boolean, Integer, Double, String, or List.")
	myDict = topologic.Dictionary.ByKeysValues(stl_keys, stl_values)
	return myDict

def mergeDictionaries(sources):
	if isinstance(sources, list) == False:
		sources = [sources]
	sinkKeys = []
	sinkValues = []
	d = sources[0].GetDictionary()
	if d != None:
		stlKeys = d.Keys()
		if len(stlKeys) > 0:
			sinkKeys = d.Keys()
			sinkValues = getValues(d)
	for i in range(1,len(sources)):
		d = sources[i].GetDictionary()
		if d == None:
			continue
		stlKeys = d.Keys()
		if len(stlKeys) > 0:
			sourceKeys = d.Keys()
			for aSourceKey in sourceKeys:
				if aSourceKey not in sinkKeys:
					sinkKeys.append(aSourceKey)
					sinkValues.append("")
			for i in range(len(sourceKeys)):
				index = sinkKeys.index(sourceKeys[i])
				sourceValue = getValueAtKey(d, sourceKeys[i])
				if sourceValue != None:
					if sinkValues[index] != "":
						if isinstance(sinkValues[index], list):
							sinkValues[index].append(sourceValue)
						else:
							sinkValues[index] = [sinkValues[index], sourceValue]
					else:
						sinkValues[index] = sourceValue
	if len(sinkKeys) > 0 and len(sinkValues) > 0:
		return processKeysValues(sinkKeys, sinkValues)
	return None

def mergeDictionaries2(sources):
	if isinstance(sources, list) == False:
		sources = [sources]
	sinkKeys = []
	sinkValues = []
	d = sources[0]
	if d != None:
		stlKeys = d.Keys()
		if len(stlKeys) > 0:
			sinkKeys = d.Keys()
			sinkValues = getValues(d)
	for i in range(1,len(sources)):
		d = sources[i]
		if d == None:
			continue
		stlKeys = d.Keys()
		if len(stlKeys) > 0:
			sourceKeys = d.Keys()
			for aSourceKey in sourceKeys:
				if aSourceKey not in sinkKeys:
					sinkKeys.append(aSourceKey)
					sinkValues.append("")
			for i in range(len(sourceKeys)):
				index = sinkKeys.index(sourceKeys[i])
				sourceValue = getValueAtKey(d, sourceKeys[i])
				if sourceValue != None:
					if sinkValues[index] != "":
						if isinstance(sinkValues[index], list):
							sinkValues[index].append(sourceValue)
						else:
							sinkValues[index] = [sinkValues[index], sourceValue]
					else:
						sinkValues[index] = sourceValue
	if len(sinkKeys) > 0 and len(sinkValues) > 0:
		return processKeysValues(sinkKeys, sinkValues)
	return None

def processCellComplex(item):
	topology, direct, directApertures, viaSharedTopologies, viaSharedApertures, toExteriorTopologies, toExteriorApertures, toContents, useInternalVertex, storeBRep, tolerance = item
	graph = None
	edges = []
	vertices = []
	cellmat = []
	if direct == True:
		cells = []
		_ = topology.Cells(None, cells)
		# Create a matrix of zeroes
		for i in range(len(cells)):
			cellRow = []
			for j in range(len(cells)):
				cellRow.append(0)
			cellmat.append(cellRow)
		for i in range(len(cells)):
			for j in range(len(cells)):
				if (i != j) and cellmat[i][j] == 0:
					cellmat[i][j] = 1
					cellmat[j][i] = 1
					sharedt = []
					cells[i].SharedTopologies(cells[j], 8, sharedt)
					if len(sharedt) > 0:
						if useInternalVertex == True:
							v1 = topologic.CellUtility.InternalVertex(cells[i], tolerance)
							v2 = topologic.CellUtility.InternalVertex(cells[j], tolerance)
						else:
							v1 = cells[i].CenterOfMass()
							v2 = cells[j].CenterOfMass()
						e = topologic.Edge.ByStartVertexEndVertex(v1, v2)
						mDict = mergeDictionaries(sharedt)
						if mDict:
							e.SetDictionary(mDict)
						edges.append(e)
	if directApertures == True:
		cellmat = []
		cells = []
		_ = topology.Cells(None, cells)
		# Create a matrix of zeroes
		for i in range(len(cells)):
			cellRow = []
			for j in range(len(cells)):
				cellRow.append(0)
			cellmat.append(cellRow)
		for i in range(len(cells)):
			for j in range(len(cells)):
				if (i != j) and cellmat[i][j] == 0:
					cellmat[i][j] = 1
					cellmat[j][i] = 1
					sharedt = []
					cells[i].SharedTopologies(cells[j], 8, sharedt)
					if len(sharedt) > 0:
						apertureExists = False
						for x in sharedt:
							apList = []
							_ = x.Apertures(apList)
							if len(apList) > 0:
								apTopList = []
								for ap in apList:
									apTopList.append(ap.Topology())
								apertureExists = True
								break
						if apertureExists:
							if useInternalVertex == True:
								v1 = topologic.CellUtility.InternalVertex(cells[i], tolerance)
								v2 = topologic.CellUtility.InternalVertex(cells[j], tolerance)
							else:
								v1 = cells[i].CenterOfMass()
								v2 = cells[j].CenterOfMass()
							e = topologic.Edge.ByStartVertexEndVertex(v1, v2)
							mDict = mergeDictionaries(apTopList)
							if mDict:
								e.SetDictionary(mDict)
							edges.append(e)

	cells = []
	_ = topology.Cells(None, cells)
	if (viaSharedTopologies == True) or (viaSharedApertures == True) or (toExteriorTopologies == True) or (toExteriorApertures == True) or (toContents == True):
		for aCell in cells:
			if useInternalVertex == True:
				vCell = topologic.CellUtility.InternalVertex(aCell, tolerance)
			else:
				vCell = aCell.CenterOfMass()
			d1 = aCell.GetDictionary()
			if storeBRep:
				d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [aCell.String(), aCell.Type(), aCell.GetTypeAsString()])
				d3 = mergeDictionaries2([d1, d2])
				_ = vCell.SetDictionary(d3)
			else:
				_ = vCell.SetDictionary(d1)
			vertices.append(vCell)
			faces = []
			_ = aCell.Faces(None, faces)
			sharedTopologies = []
			exteriorTopologies = []
			sharedApertures = []
			exteriorApertures = []
			contents = []
			_ = aCell.Contents(contents)
			for aFace in faces:
				cells = []
				_ = aFace.Cells(topology, cells)
				if len(cells) > 1:
					sharedTopologies.append(aFace)
					apertures = []
					_ = aFace.Apertures(apertures)
					for anAperture in apertures:
						sharedApertures.append(anAperture)
				else:
					exteriorTopologies.append(aFace)
					apertures = []
					_ = aFace.Apertures(apertures)
					for anAperture in apertures:
						exteriorApertures.append(anAperture)
			if viaSharedTopologies:
				for sharedTopology in sharedTopologies:
					if useInternalVertex == True:
						vst = internalVertex(sharedTopology, tolerance)
					else:
						vst = sharedTopology.CenterOfMass()
					d1 = sharedTopology.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [sharedTopology.String(), sharedTopology.Type(), sharedTopology.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vCell, vst)
					tempd = processKeysValues(["relationship"],["Via Shared Topologies"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
					if toContents:
						contents = []
						_ = sharedTopology.Contents(contents)
						for content in contents:
							if useInternalVertex == True:
								vst2 = internalVertex(content, tolerance)
							else:
								vst2 = content.CenterOfMass()
							d1 = content.GetDictionary()
							vst2 = topologic.Vertex.ByCoordinates(vst2.X(), vst2.Y(), vst2.Z())
							if storeBRep:
								d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
								d3 = mergeDictionaries2([d1, d2])
								_ = vst2.SetDictionary(d3)
							else:
								_ = vst2.SetDictionary(d1)
							vertices.append(vst2)
							tempe = topologic.Edge.ByStartVertexEndVertex(vst, vst2)
							tempd = processKeysValues(["relationship"],["To Contents"])
							_ = tempe.SetDictionary(tempd)
							edges.append(tempe)
			if viaSharedApertures:
				for sharedAperture in sharedApertures:
					if useInternalVertex == True:
						vst = internalVertex(sharedAperture.Topology(), tolerance)
					else:
						vst = sharedAperture.Topology().CenterOfMass()
					d1 = sharedAperture.Topology().GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [sharedAperture.Topology().String(), sharedAperture.Topology().Type(), sharedAperture.Topology().GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vCell, vst)
					tempd = processKeysValues(["relationship"],["Via Shared Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
			if toExteriorTopologies:
				for exteriorTopology in exteriorTopologies:
					if useInternalVertex == True:
						vst = internalVertex(exteriorTopology, tolerance)
					else:
						vst = exteriorTopology.CenterOfMass()
					_ = vst.SetDictionary(exteriorTopology.GetDictionary())
					d1 = exteriorTopology.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorTopology.String(), exteriorTopology.Type(), exteriorTopology.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vCell, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Topologies"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
					if toContents:
						contents = []
						_ = exteriorTopology.Contents(contents)
						for content in contents:
							if useInternalVertex == True:
								vst2 = internalVertex(content, tolerance)
							else:
								vst2 = content.CenterOfMass()
							d1 = content.GetDictionary()
							vst2 = topologic.Vertex.ByCoordinates(vst2.X()+(tolerance*100), vst2.Y()+(tolerance*100), vst2.Z()+(tolerance*100))
							if storeBRep:
								d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
								d3 = mergeDictionaries2([d1, d2])
								_ = vst2.SetDictionary(d3)
							else:
								_ = vst2.SetDictionary(d1)
							vertices.append(vst2)
							tempe = topologic.Edge.ByStartVertexEndVertex(vst, vst2)
							tempd = processKeysValues(["relationship"],["To Contents"])
							_ = tempe.SetDictionary(tempd)
							edges.append(tempe)
			if toExteriorApertures:
				for exteriorAperture in exteriorApertures:
					extTop = exteriorAperture.Topology()
					if useInternalVertex == True:
						vst = internalVertex(extTop, tolerance)
					else:
						vst = exteriorAperture.Topology().CenterOfMass()
					d1 = exteriorAperture.Topology().GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorAperture.String(), exteriorAperture.Type(), exteriorAperture.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vCell, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
			if toContents:
				contents = []
				_ = aCell.Contents(contents)
				for content in contents:
					if useInternalVertex == True:
						vst = internalVertex(content, tolerance)
					else:
						vst = content.CenterOfMass()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					d1 = content.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vCell, vst)
					tempd = processKeysValues(["relationship"],["To Contents"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)

	for aCell in cells:
		if useInternalVertex == True:
			vCell = topologic.CellUtility.InternalVertex(aCell, tolerance)
		else:
			vCell = aCell.CenterOfMass()
		d1 = aCell.GetDictionary()
		if storeBRep:
			d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [aCell.String(), aCell.Type(), aCell.GetTypeAsString()])
			d3 = mergeDictionaries2([d1, d2])
			_ = vCell.SetDictionary(d3)
		else:
			_ = vCell.SetDictionary(d1)
		vertices.append(vCell)
	return topologic.Graph.ByVerticesEdges(vertices,edges)

def processCell(item):
	topology, direct, directApertures, viaSharedTopologies, viaSharedApertures, toExteriorTopologies, toExteriorApertures, toContents, useInternalVertex, storeBRep, tolerance = item
	vertices = []
	edges = []

	if useInternalVertex == True:
		vCell = topologic.CellUtility.InternalVertex(topology, tolerance)
	else:
		vCell = topology.CenterOfMass()
	d1 = topology.GetDictionary()
	if storeBRep:
		d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [topology.String(), topology.Type(), topology.GetTypeAsString()])
		d3 = mergeDictionaries2([d1, d2])
		_ = vCell.SetDictionary(d3)
	else:
		_ = vCell.SetDictionary(d1)
	vertices.append(vCell)

	if (toExteriorTopologies == True) or (toExteriorApertures == True) or (toContents == True):
		faces = []
		_ = topology.Faces(None, faces)
		exteriorTopologies = []
		exteriorApertures = []
		for aFace in faces:
			exteriorTopologies.append(aFace)
			apertures = []
			_ = aFace.Apertures(apertures)
			for anAperture in apertures:
				exteriorApertures.append(anAperture)
			if toExteriorTopologies:
				for exteriorTopology in exteriorTopologies:
					if useInternalVertex == True:
						vst = internalVertex(exteriorTopology, tolerance)
					else:
						vst = exteriorTopology.CenterOfMass()
					d1 = exteriorTopology.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorTopology.String(), exteriorTopology.Type(), exteriorTopology.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vCell, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Topologies"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
					if toContents:
						contents = []
						_ = exteriorTopology.Contents(contents)
						for content in contents:
							if useInternalVertex == True:
								vst2 = internalVertex(content, tolerance)
							else:
								vst2 = content.CenterOfMass()
							vst2 = topologic.Vertex.ByCoordinates(vst2.X()+(tolerance*100), vst2.Y()+(tolerance*100), vst2.Z()+(tolerance*100))
							d1 = content.GetDictionary()
							if storeBRep:
								d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
								d3 = mergeDictionaries2([d1, d2])
								_ = vst2.SetDictionary(d3)
							else:
								_ = vst2.SetDictionary(d1)
							vertices.append(vst2)
							tempe = topologic.Edge.ByStartVertexEndVertex(vst, vst2)
							tempd = processKeysValues(["relationship"],["To Contents"])
							_ = tempe.SetDictionary(tempd)
							edges.append(tempe)
			if toExteriorApertures:
				for exteriorAperture in exteriorApertures:
					extTop = exteriorAperture.Topology()
					if useInternalVertex == True:
						vst = internalVertex(extTop, tolerance)
					else:
						vst = exteriorAperture.Topology().CenterOfMass()
					d1 = exteriorAperture.Topology().GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorAperture.Topology().String(), exteriorAperture.Topology().Type(), exteriorAperture.Topology().GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vCell, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
			if toContents:
				contents = []
				_ = topology.Contents(contents)
				for content in contents:
					if useInternalVertex == True:
						vst = internalVertex(content, tolerance)
					else:
						vst = content.CenterOfMass()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					d1 = content.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vCell, vst)
					tempd = processKeysValues(["relationship"],["To Contents"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)

	return topologic.Graph.ByVerticesEdges(vertices, edges)

def processShell(item):
	topology, direct, directApertures, viaSharedTopologies, viaSharedApertures, toExteriorTopologies, toExteriorApertures, toContents, useInternalVertex, storeBRep, tolerance = item
	graph = None
	edges = []
	vertices = []
	facemat = []
	if direct == True:
		topFaces = []
		_ = topology.Faces(None, topFaces)
		# Create a matrix of zeroes
		for i in range(len(topFaces)):
			faceRow = []
			for j in range(len(topFaces)):
				faceRow.append(0)
			facemat.append(faceRow)
		for i in range(len(topFaces)):
			for j in range(len(topFaces)):
				if (i != j) and facemat[i][j] == 0:
					facemat[i][j] = 1
					facemat[j][i] = 1
					sharedt = []
					topFaces[i].SharedTopologies(topFaces[j], 2, sharedt)
					if len(sharedt) > 0:
						if useInternalVertex == True:
							v1 = topologic.FaceUtility.InternalVertex(topFaces[i], tolerance)
							v2 = topologic.FaceUtility.InternalVertex(topFaces[j], tolerance)
						else:
							v1 = topFaces[i].CenterOfMass()
							v2 = topFaces[j].CenterOfMass()
						e = topologic.Edge.ByStartVertexEndVertex(v1, v2)
						mDict = mergeDictionaries(sharedt)
						if mDict:
							e.SetDictionary(mDict)
						edges.append(e)
	if directApertures == True:
		facemat = []
		topFaces = []
		_ = topology.Faces(None, topFaces)
		# Create a matrix of zeroes
		for i in range(len(topFaces)):
			faceRow = []
			for j in range(len(topFaces)):
				faceRow.append(0)
			facemat.append(faceRow)
		for i in range(len(topFaces)):
			for j in range(len(topFaces)):
				if (i != j) and facemat[i][j] == 0:
					facemat[i][j] = 1
					facemat[j][i] = 1
					sharedt = []
					topFaces[i].SharedTopologies(topFaces[j], 2, sharedt)
					if len(sharedt) > 0:
						apertureExists = False
						for x in sharedt:
							apList = []
							_ = x.Apertures(apList)
							if len(apList) > 0:
								apertureExists = True
								break
						if apertureExists:
							apTopList = []
							for ap in apList:
								apTopList.append(ap.Topology())
							if useInternalVertex == True:
								v1 = topologic.FaceUtility.InternalVertex(topFaces[i], tolerance)
								v2 = topologic.FaceUtility.InternalVertex(topFaces[j], tolerance)
							else:
								v1 = topFaces[i].CenterOfMass()
								v2 = topFaces[j].CenterOfMass()
							e = topologic.Edge.ByStartVertexEndVertex(v1, v2)
							mDict = mergeDictionaries(apTopList)
							if mDict:
								e.SetDictionary(mDict)
							edges.append(e)

	topFaces = []
	_ = topology.Faces(None, topFaces)
	if (viaSharedTopologies == True) or (viaSharedApertures == True) or (toExteriorTopologies == True) or (toExteriorApertures == True) or (toContents == True):
		for aFace in topFaces:
			if useInternalVertex == True:
				vFace = topologic.FaceUtility.InternalVertex(aFace, tolerance)
			else:
				vFace = aFace.CenterOfMass()
			_ = vFace.SetDictionary(aFace.GetDictionary())
			vertices.append(vFace)
			fEdges = []
			_ = aFace.Edges(None, fEdges)
			sharedTopologies = []
			exteriorTopologies = []
			sharedApertures = []
			exteriorApertures = []
			for anEdge in fEdges:
				faces = []
				_ = anEdge.Faces(topology, faces)
				if len(faces) > 1:
					sharedTopologies.append(anEdge)
					apertures = []
					_ = anEdge.Apertures(apertures)
					for anAperture in apertures:
						sharedApertures.append(anAperture)
				else:
					exteriorTopologies.append(anEdge)
					apertures = []
					_ = anEdge.Apertures(apertures)
					for anAperture in apertures:
						exteriorApertures.append(anAperture)
			if viaSharedTopologies:
				for sharedTopology in sharedTopologies:
					if useInternalVertex == True:
						vst = internalVertex(sharedTopology, tolerance)
					else:
						vst = sharedTopology.CenterOfMass()
					d1 = sharedTopology.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [sharedTopology.String(), sharedTopology.Type(), sharedTopology.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vFace, vst)
					tempd = processKeysValues(["relationship"],["Via Shared Topologies"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
					if toContents:
						contents = []
						_ = sharedTopology.Contents(contents)
						for content in contents:
							if useInternalVertex == True:
								vst2 = internalVertex(content, tolerance)
							else:
								vst2 = content.CenterOfMass()
							vst2 = topologic.Vertex.ByCoordinates(vst2.X()+(tolerance*100), vst2.Y()+(tolerance*100), vst2.Z()+(tolerance*100))
							d1 = content.GetDictionary()
							if storeBRep:
								d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
								d3 = mergeDictionaries2([d1, d2])
								_ = vst2.SetDictionary(d3)
							else:
								_ = vst2.SetDictionary(d1)
							vertices.append(vst2)
							tempe = topologic.Edge.ByStartVertexEndVertex(vst, vst2)
							tempd = processKeysValues(["relationship"],["To Contents"])
							_ = tempe.SetDictionary(tempd)
							edges.append(tempe)
			if viaSharedApertures:
				for sharedAperture in sharedApertures:
					if useInternalVertex == True:
						vst = internalVertex(sharedAperture.Topology(), tolerance)
					else:
						vst = sharedAperture.Topology().CenterOfMass()
					d1 = sharedAperture.Topology().GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [sharedAperture.Topology().String(), sharedAperture.Topology().Type(), sharedAperture.Topology().GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vFace, vst)
					tempd = processKeysValues(["relationship"],["Via Shared Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
			if toExteriorTopologies:
				for exteriorTopology in exteriorTopologies:
					if useInternalVertex == True:
						vst = internalVertex(exteriorTopology, tolerance)
					else:
						vst = exteriorTopology.CenterOfMass()
					d1 = exteriorTopology.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorTopology.String(), exteriorTopology.Type(), exteriorTopology.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vFace, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
					if toContents:
						contents = []
						_ = exteriorTopology.Contents(contents)
						for content in contents:
							if useInternalVertex == True:
								vst2 = internalVertex(content, tolerance)
							else:
								vst2 = content.CenterOfMass()
							vst2 = topologic.Vertex.ByCoordinates(vst2.X()+(tolerance*100), vst2.Y()+(tolerance*100), vst2.Z()+(tolerance*100))
							d1 = content.GetDictionary()
							if storeBRep:
								d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
								d3 = mergeDictionaries2([d1, d2])
								_ = vst2.SetDictionary(d3)
							else:
								_ = vst2.SetDictionary(d1)
							vertices.append(vst2)
							tempe = topologic.Edge.ByStartVertexEndVertex(vst, vst2)
							tempd = processKeysValues(["relationship"],["To Contents"])
							_ = tempe.SetDictionary(tempd)
							edges.append(tempe)
			if toExteriorApertures:
				for exteriorAperture in exteriorApertures:
					extTop = exteriorAperture.Topology()
					if useInternalVertex == True:
						vst = internalVertex(extTop, tolerance)
					else:
						vst = exteriorAperture.Topology().CenterOfMass()
					d1 = exteriorAperture.Topology().GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorAperture.Topology().String(), exteriorAperture.Topology().Type(), exteriorAperture.Topology().GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vFace, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
			if toContents:
				contents = []
				_ = aFace.Contents(contents)
				for content in contents:
					if useInternalVertex == True:
						vst = internalVertex(content, tolerance)
					else:
						vst = content.CenterOfMass()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					d1 = content.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vFace, vst)
					tempd = processKeysValues(["relationship"],["To Contents"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)

	for aFace in topFaces:
		if useInternalVertex == True:
			vFace = internalVertex(aFace, tolerance)
		else:
			vFace = aFace.CenterOfMass()
		d1 = aFace.GetDictionary()
		if storeBRep:
			d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [aFace.String(), aFace.Type(), aFace.GetTypeAsString()])
			d3 = mergeDictionaries2([d1, d2])
			_ = vFace.SetDictionary(d3)
		else:
			_ = vFace.SetDictionary(d1)
		vertices.append(vFace)
	return topologic.Graph.ByVerticesEdges(vertices, edges)

def processFace(item):
	topology, direct, directApertures, viaSharedTopologies, viaSharedApertures, toExteriorTopologies, toExteriorApertures, toContents, useInternalVertex, storeBRep, tolerance = item

	graph = None
	vertices = []
	edges = []

	if useInternalVertex == True:
		vFace = topologic.FaceUtility.InternalVertex(topology, tolerance)
	else:
		vFace = topology.CenterOfMass()
	d1 = topology.GetDictionary()
	if storeBRep:
		d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [topology.String(), topology.Type(), topology.GetTypeAsString()])
		d3 = mergeDictionaries2([d1, d2])
		_ = vFace.SetDictionary(d3)
	else:
		_ = vFace.SetDictionary(d1)
	vertices.append(vFace)
	if (toExteriorTopologies == True) or (toExteriorApertures == True) or (toContents == True):
		fEdges = []
		_ = topology.Edges(None, fEdges)
		exteriorTopologies = []
		exteriorApertures = []

		for anEdge in fEdges:
			exteriorTopologies.append(anEdge)
			apertures = []
			_ = anEdge.Apertures(apertures)
			for anAperture in apertures:
				exteriorApertures.append(anAperture)
			if toExteriorTopologies:
				for exteriorTopology in exteriorTopologies:
					if useInternalVertex == True:
						vst = internalVertex(exteriorTopology, tolerance)
					else:
						vst = exteriorTopology.CenterOfMass()
					d1 = exteriorTopology.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorTopology.String(), exteriorTopology.Type(), exteriorTopology.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vFace, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Topologies"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
					if toContents:
						contents = []
						_ = exteriorTopology.Contents(contents)
						for content in contents:
							if useInternalVertex == True:
								vst2 = internalVertex(content, tolerance)
							else:
								vst2 = content.CenterOfMass()
							vst2 = topologic.Vertex.ByCoordinates(vst2.X()+(tolerance*100), vst2.Y()+(tolerance*100), vst2.Z()+(tolerance*100))
							d1 = content.GetDictionary()
							if storeBRep:
								d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
								d3 = mergeDictionaries2([d1, d2])
								_ = vst2.SetDictionary(d3)
							else:
								_ = vst2.SetDictionary(d1)
							vertices.append(vst2)
							tempe = topologic.Edge.ByStartVertexEndVertex(vst, vst2)
							tempd = processKeysValues(["relationship"],["To Contents"])
							_ = tempe.SetDictionary(tempd)
							edges.append(tempe)
			if toExteriorApertures:
				for exteriorAperture in exteriorApertures:
					extTop = exteriorAperture.Topology()
					if useInternalVertex == True:
						vst = internalVertex(extTop, tolerance)
					else:
						vst = exteriorAperture.Topology().CenterOfMass()
					d1 = exteriorAperture.Topology().GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorAperture.Topology().String(), exteriorAperture.Topology().Type(), exteriorAperture.Topology().GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vFace, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
			if toContents:
				contents = []
				_ = topology.Contents(contents)
				for content in contents:
					if useInternalVertex == True:
						vst = internalVertex(content, tolerance)
					else:
						vst = content.CenterOfMass()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					d1 = content.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vFace, vst)
					tempd = processKeysValues(["relationship"],["To Contents"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
	return topologic.Graph.ByVertices(vertices, edges)

def processWire(item):
	topology, direct, directApertures, viaSharedTopologies, viaSharedApertures, toExteriorTopologies, toExteriorApertures, toContents, useInternalVertex, storeBRep, tolerance = item
	graph = None
	edges = []
	vertices = []
	edgemat = []
	if direct == True:
		topEdges = []
		_ = topology.Edges(None, topEdges)
		# Create a matrix of zeroes
		for i in range(len(topEdges)):
			edgeRow = []
			for j in range(len(topEdges)):
				edgeRow.append(0)
			edgemat.append(edgeRow)
		for i in range(len(topEdges)):
			for j in range(len(topEdges)):
				if (i != j) and edgemat[i][j] == 0:
					edgemat[i][j] = 1
					edgemat[j][i] = 1
					sharedt = []
					topEdges[i].SharedTopologies(topEdges[j], 1, sharedt)
					if len(sharedt) > 0:
						try:
							v1 = topologic.EdgeUtility.PointAtParameter(topEdges[i], 0.5)
						except:
							v1 = topEdges[j].CenterOfMass()
						try:
							v2 = topologic.EdgeUtility.PointAtParameter(topEdges[j], 0.5)
						except:
							v2 = topEdges[j].CenterOfMass()
						e = topologic.Edge.ByStartVertexEndVertex(v1, v2)
						mDict = mergeDictionaries(sharedt)
						if mDict:
							e.SetDictionary(mDict)
						edges.append(e)
	if directApertures == True:
		edgemat = []
		topEdges = []
		_ = topology.Edges(None, topEdges)
		# Create a matrix of zeroes
		for i in range(len(topEdges)):
			edgeRow = []
			for j in range(len(topEdges)):
				edgeRow.append(0)
			edgemat.append(edgeRow)
		for i in range(len(topEdges)):
			for j in range(len(topEdges)):
				if (i != j) and edgemat[i][j] == 0:
					edgemat[i][j] = 1
					edgemat[j][i] = 1
					sharedt = []
					topEdges[i].SharedTopologies(topEdges[j], 1, sharedt)
					if len(sharedt) > 0:
						apertureExists = False
						for x in sharedt:
							apList = []
							_ = x.Apertures(apList)
							if len(apList) > 0:
								apertureExists = True
								break
						if apertureExists:
							try:
								v1 = topologic.EdgeUtility.PointAtParameter(topEdges[i], 0.5)
							except:
								v1 = topEdges[j].CenterOfMass()
							try:
								v2 = topologic.EdgeUtility.PointAtParameter(topEdges[j], 0.5)
							except:
								v2 = topEdges[j].CenterOfMass()
							e = topologic.Edge.ByStartVertexEndVertex(v1, v2)
							apTopologies = []
							for ap in apList:
								apTopologies.append(ap.Topology())
							mDict = mergeDictionaries(apTopologies)
							if mDict:
								e.SetDictionary(mDict)
							edges.append(e)

	topEdges = []
	_ = topology.Edges(None, topEdges)
	if (viaSharedTopologies == True) or (viaSharedApertures == True) or (toExteriorTopologies == True) or (toExteriorApertures == True) or (toContents == True):
		for anEdge in topEdges:
			try:
				vEdge = topologic.EdgeUtility.PointAtParameter(anEdge, 0.5)
			except:
				vEdge = anEdge.CenterOfMass()
			d1 = anEdge.GetDictionary()
			if storeBRep:
				d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [anEdge.String(), anEdge.Type(), anEdge.GetTypeAsString()])
				d3 = mergeDictionaries2([d1, d2])
				_ = vEdge.SetDictionary(d3)
			else:
				_ = vEdge.SetDictionary(d1)
			vertices.append(vEdge)
			eVertices = []
			_ = anEdge.Vertices(None, eVertices)
			sharedTopologies = []
			exteriorTopologies = []
			sharedApertures = []
			exteriorApertures = []
			contents = []
			_ = anEdge.Contents(contents)
			for aVertex in eVertices:
				tempEdges = []
				_ = aVertex.Edges(topology, tempEdges)
				if len(tempEdges) > 1:
					sharedTopologies.append(aVertex)
					apertures = []
					_ = aVertex.Apertures(apertures)
					for anAperture in apertures:
						sharedApertures.append(anAperture)
				else:
					exteriorTopologies.append(aVertex)
					apertures = []
					_ = aVertex.Apertures(apertures)
					for anAperture in apertures:
						exteriorApertures.append(anAperture)
			if viaSharedTopologies:
				for sharedTopology in sharedTopologies:
					vst = sharedTopology.CenterOfMass()
					d1 = sharedTopology.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [sharedTopology.String(), sharedTopology.Type(), sharedTopology.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vEdge, vst)
					tempd = processKeysValues(["relationship"],["Via Shared Topologies"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
					if toContents:
						contents = []
						_ = sharedTopology.Contents(contents)
						for content in contents:
							if useInternalVertex == True:
								vst2 = internalVertex(content, tolerance)
							else:
								vst2 = content.CenterOfMass()
							vst2 = topologic.Vertex.ByCoordinates(vst2.X()+(tolerance*100), vst2.Y()+(tolerance*100), vst2.Z()+(tolerance*100))
							d1 = content.GetDictionary()
							if storeBRep:
								d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
								d3 = mergeDictionaries2([d1, d2])
								_ = vst2.SetDictionary(d3)
							else:
								_ = vst2.SetDictionary(d1)
							vertices.append(vst2)
							tempe = topologic.Edge.ByStartVertexEndVertex(vst, vst2)
							tempd = processKeysValues(["relationship"],["To Contents"])
							_ = tempe.SetDictionary(tempd)
							edges.append(tempe)
			if viaSharedApertures:
				for sharedAperture in sharedApertures:
					if useInternalVertex == True:
						vst = internalVertex(sharedAperture.Topology(), tolerance)
					else:
						vst = sharedAperture.Topology().CenterOfMass()
					d1 = sharedAperture.Topology().GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [sharedAperture.Topology().String(), sharedAperture.Topology().Type(), sharedAperture.Topology().GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vEdge, vst)
					tempd = processKeysValues(["relationship"],["Via Shared Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
			if toExteriorTopologies:
				for exteriorTopology in exteriorTopologies:
					vst = exteriorTopology
					vertices.append(exteriorTopology)
					tempe = topologic.Edge.ByStartVertexEndVertex(vEdge, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Topologies"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
					if toContents:
						contents = []
						_ = vst.Contents(contents)
						for content in contents:
							if useInternalVertex == True:
								vst2 = internalVertex(content, tolerance)
							else:
								vst2 = content.CenterOfMass()
							vst2 = topologic.Vertex.ByCoordinates(vst2.X()+(tolerance*100), vst2.Y()+(tolerance*100), vst2.Z()+(tolerance*100))
							d1 = content.GetDictionary()
							if storeBRep:
								d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
								d3 = mergeDictionaries2([d1, d2])
								_ = vst2.SetDictionary(d3)
							else:
								_ = vst2.SetDictionary(d1)
							vertices.append(vst2)
							tempe = topologic.Edge.ByStartVertexEndVertex(vst, vst2)
							tempd = processKeysValues(["relationship"],["To Contents"])
							_ = tempe.SetDictionary(tempd)
							edges.append(tempe)
			if toExteriorApertures:
				for exteriorAperture in exteriorApertures:
					extTop = exteriorAperture.Topology()
					if useInternalVertex == True:
						vst = internalVertex(extTop, tolerance)
					else:
						vst = extTop.CenterOfMass()
					d1 = extTop.GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [extTop.String(), extTop.Type(), extTop.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vEdge, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
			if toContents:
				contents = []
				_ = anEdge.Contents(contents)
				for content in contents:
					if useInternalVertex == True:
						vst = internalVertex(content, tolerance)
					else:
						vst = content.CenterOfMass()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					d1 = content.GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X(), vst.Y(), vst.Z())
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vEdge, vst)
					tempd = processKeysValues(["relationship"],["To Contents"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
	for anEdge in topEdges:
		try:
			vEdge = topologic.EdgeUtility.PointAtParameter(anEdge, 0.5)
		except:
			vEdge = anEdge.CenterOfMass()
		d1 = anEdge.GetDictionary()
		if storeBRep:
			d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [anEdge.String(), anEdge.Type(), anEdge.GetTypeAsString()])
			d3 = mergeDictionaries2([d1, d2])
			_ = vEdge.SetDictionary(d3)
		else:
			_ = vEdge.SetDictionary(d1)
		vertices.append(vEdge)
	return topologic.Graph.ByVerticesEdges(vertices, edges)

def processEdge(item):
	topology, direct, directApertures, viaSharedTopologies, viaSharedApertures, toExteriorTopologies, toExteriorApertures, toContents, useInternalVertex, storeBRep, tolerance = item
	graph = None
	vertices = []
	edges = []

	if useInternalVertex == True:
		try:
			vEdge = topologic.EdgeUtility.PointAtParameter(topology, 0.5)
		except:
			vEdge = topology.CenterOfMass()
	else:
		vEdge = topology.CenterOfMass()

	d1 = vEdge.GetDictionary()
	if storeBRep:
		d2 = processKeysValues(["brep"], [topology.String()])
		d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [topology.String(), topology.Type(), topology.GetTypeAsString()])
		d3 = mergeDictionaries2([d1, d2])
		_ = vEdge.SetDictionary(d3)
	else:
		_ = vEdge.SetDictionary(topology.GetDictionary())

	vertices.append(vEdge)

	if (toExteriorTopologies == True) or (toExteriorApertures == True) or (toContents == True):
		eVertices = []
		_ = topology.Vertices(None, eVertices)
		exteriorTopologies = []
		exteriorApertures = []
		for aVertex in eVertices:
			exteriorTopologies.append(aVertex)
			apertures = []
			_ = aVertex.Apertures(apertures)
			for anAperture in apertures:
				exteriorApertures.append(anAperture)
			if toExteriorTopologies:
				for exteriorTopology in exteriorTopologies:
					if useInternalVertex == True:
						vst = internalVertex(exteriorTopology, tolerance)
					else:
						vst = exteriorTopology.CenterOfMass()
					d1 = exteriorTopology.GetDictionary()
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorTopology.String(), exteriorTopology.Type(), exteriorTopology.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vEdge, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Topologies"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
					if toContents:
						contents = []
						_ = vst.Contents(contents)
						for content in contents:
							if useInternalVertex == True:
								vst2 = internalVertex(content, tolerance)
							else:
								vst2 = content.CenterOfMass()
							vst2 = topologic.Vertex.ByCoordinates(vst2.X()+(tolerance*100), vst2.Y()+(tolerance*100), vst2.Z()+(tolerance*100))
							d1 = content.GetDictionary()
							if storeBRep:
								d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
								d3 = mergeDictionaries2([d1, d2])
								_ = vst2.SetDictionary(d3)
							else:
								_ = vst2.SetDictionary(d1)
							vertices.append(vst2)
							tempe = topologic.Edge.ByStartVertexEndVertex(vst, vst2)
							tempd = processKeysValues(["relationship"],["To Contents"])
							_ = tempe.SetDictionary(tempd)
							edges.append(tempe)
			if toExteriorApertures:
				for exteriorAperture in exteriorApertures:
					extTop = exteriorAperture.Topology()
					if useInternalVertex == True:
						vst = internalVertex(extTop, tolerance)
					else:
						vst = exteriorAperture.Topology().CenterOfMass()
					d1 = exteriorAperture.Topology().GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [exteriorAperture.Topology().String(), exteriorAperture.Topology().Type(), exteriorAperture.Topology().GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					_ = vst.SetDictionary(exteriorAperture.Topology().GetDictionary())
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vEdge, vst)
					tempd = processKeysValues(["relationship"],["To Exterior Apertures"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
			if toContents:
				contents = []
				_ = topology.Contents(contents)
				for content in contents:
					if useInternalVertex == True:
						vst = internalVertex(content, tolerance)
					else:
						vst = content.CenterOfMass()
					d1 = content.GetDictionary()
					vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
					if storeBRep:
						d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
						d3 = mergeDictionaries2([d1, d2])
						_ = vst.SetDictionary(d3)
					else:
						_ = vst.SetDictionary(d1)
					vertices.append(vst)
					tempe = topologic.Edge.ByStartVertexEndVertex(vEdge, vst)
					tempd = processKeysValues(["relationship"],["To Contents"])
					_ = tempe.SetDictionary(tempd)
					edges.append(tempe)
	graph = topologic.Graph.ByVerticesEdges(vertices, edges)
	return graph

def processVertex(item):
	topology, direct, directApertures, viaSharedTopologies, viaSharedApertures, toExteriorTopologies, toExteriorApertures, toContents, useInternalVertex, storeBRep, tolerance = item
	vertices = [topology]
	edges = []
	if toContents:
		contents = []
		_ = topology.Contents(contents)
		for content in contents:
			if useInternalVertex == True:
				vst = internalVertex(content, tolerance)
			else:
				vst = content.CenterOfMass()
			d1 = content.GetDictionary()
			vst = topologic.Vertex.ByCoordinates(vst.X()+(tolerance*100), vst.Y()+(tolerance*100), vst.Z()+(tolerance*100))
			if storeBRep:
				d2 = processKeysValues(["brep", "brepType", "brepTypeString"], [content.String(), content.Type(), content.GetTypeAsString()])
				d3 = mergeDictionaries2([d1, d2])
				_ = vst.SetDictionary(d3)
			else:
				_ = vst.SetDictionary(d1)
			vertices.append(vst)
			tempe = topologic.Edge.ByStartVertexEndVertex(topology, vst)
			tempd = processKeysValues(["relationship"],["To Contents"])
			_ = tempe.SetDictionary(tempd)
			edges.append(tempe)
	return topologic.Graph.VerticesEdges(vertices, edges)


import topologicpy
import topologic
from topologicpy.Face import Face
from topologicpy.Topology import Topology
import collections

class Vertex(Topology):
    @staticmethod
    def AreIpsilateral(vertices: list, face: topologic.Face) -> bool:
        """
        Returns True if the two input vertices are on the same side of the input face. Returns False otherwise. If at least one of the vertices is on the face, this method return True.

        Parameters
        ----------
        vertices : list
            The input list of vertices.
        face : topologic.Face
            The input face

        Returns
        -------
        bool
            True if the input vertices are on the same side of the face. False otherwise. If at least one of the vertices is on the face, this method return True.

        """
        def check(dot_productA, pointB, pointC, normal):
            # Calculate the dot products of the vectors from the surface point to each of the input points.
            dot_productB = (pointB[0] - pointC[0]) * normal[0] + \
                        (pointB[1] - pointC[1]) * normal[1] + \
                        (pointB[2] - pointC[2]) * normal[2]

            # Check if both points are on the same side of the surface.
            if dot_productA * dot_productB > 0:
                return True

            # Check if both points are on opposite sides of the surface.
            elif dot_productA * dot_productB < 0:
                return False

            # Otherwise, at least one point is on the surface.
            else:
                return True
    
        from topologicpy.Vertex import Vertex
        from topologicpy.Face import Face

        if not isinstance(face, topologic.Face):
            return None
        vertexList = [x for x in vertices if isinstance(x, topologic.Vertex)]
        if len(vertexList) < 2:
            return None
        pointA = Vertex.Coordinates(vertexList[0])
        pointC = Vertex.Coordinates(Face.VertexByParameters(face, 0.5, 0.5))
        normal = Face.Normal(face)
        dot_productA = (pointA[0] - pointC[0]) * normal[0] + \
                        (pointA[1] - pointC[1]) * normal[1] + \
                        (pointA[2] - pointC[2]) * normal[2]
        for i in range(1, len(vertexList)):
            pointB = Vertex.Coordinates(vertexList[i])
            if not check(dot_productA, pointB, pointC, normal):
                return False
        return True
    
    @staticmethod
    def AreIpsilateralCluster(cluster: topologic.Cluster, face: topologic.Face) -> bool:
        """
        Returns True if the two input vertices are on the same side of the input face. Returns False otherwise. If at least one of the vertices is on the face, this method return True.

        Parameters
        ----------
        cluster : topologic.Cluster
            The input list of vertices.
        face : topologic.Face
            The input face

        Returns
        -------
        bool
            True if the input vertices are on the same side of the face. False otherwise. If at least one of the vertices is on the face, this method return True.

        """
        from topologicpy.Topology import Topology
        if not isinstance(cluster, topologic.Topology):
            return None
        vertices = Topology.SubTopologies(cluster, subTopologyType="vertex")
        return Vertex.AreIpsilateral(vertices, face)
    
    @staticmethod
    def AreOnSameSide(vertices: list, face: topologicpy.Face.Face) -> bool:
        """
        Returns True if the two input vertices are on the same side of the input face. Returns False otherwise. If at least one of the vertices is on the face, this method return True.

        Parameters
        ----------
        vertices : list
            The input list of vertices.
        face : topologic.Face
            The input face

        Returns
        -------
        bool
            True if the input vertices are on the same side of the face. False otherwise. If at least one of the vertices is on the face, this method return True.

        """
        return Vertex.AreIpsilateral(vertices, face)

    @staticmethod
    def AreOnSameSideCluster(cluster: topologic.Cluster, face: topologic.Face) -> bool:
        """
        Returns True if the two input vertices are on the same side of the input face. Returns False otherwise. If at least one of the vertices is on the face, this method return True.

        Parameters
        ----------
        cluster : topologic.Cluster
            The input list of vertices.
        face : topologic.Face
            The input face

        Returns
        -------
        bool
            True if the input vertices are on the same side of the face. False otherwise. If at least one of the vertices is on the face, this method return True.

        """
        from topologicpy.Topology import Topology
        if not isinstance(cluster, topologic.Topology):
            return None
        vertices = Topology.SubTopologies(cluster, subTopologyType="vertex")
        return Vertex.AreIpsilateral(vertices, face)

    @staticmethod
    def ByCoordinates(x: float = 0, y: float = 0, z: float = 0) -> topologic.Vertex:
        """
        Creates a vertex at the coordinates specified by the x, y, z inputs.

        Parameters
        ----------
        x : float , optional
            The X coordinate. The default is 0.
        y : float , optional
            The Y coordinate. The default is 0.
        z : float , optional
            The Z coordinate. The defaults is 0.

        Returns
        -------
        topologic.Vertex
            The created vertex.

        """
        vertex = None
        try:
            vertex = topologic.Vertex.ByCoordinates(x, y, z)
        except:
            vertex = None
        return vertex
    
    @staticmethod
    def Coordinates(vertex: topologic.Vertex, outputType: str = "xyz", mantissa: int = 4) -> list:
        """
        Returns the coordinates of the input vertex.

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex.
        outputType : string, optional
            The desired output type. Could be any permutation or substring of "xyz" or the string "matrix". The default is "xyz". The input is case insensitive and the coordinates will be returned in the specified order.
        mantissa : int , optional
            The desired length of the mantissa. The default is 4.

        Returns
        -------
        list
            The coordinates of the input vertex.

        """
        if not isinstance(vertex, topologic.Vertex):
            return None
        x = round(vertex.X(), mantissa)
        y = round(vertex.Y(), mantissa)
        z = round(vertex.Z(), mantissa)
        matrix = [[1,0,0,x],
                [0,1,0,y],
                [0,0,1,z],
                [0,0,0,1]]
        output = []
        outputType = outputType.lower()
        if outputType == "matrix":
            return matrix
        else:
            outputType = list(outputType)
            for axis in outputType:
                if axis == "x":
                    output.append(x)
                elif axis == "y":
                    output.append(y)
                elif axis == "z":
                    output.append(z)
        return output

    @staticmethod
    def Distance(vertex: topologic.Vertex, topology: topologic.Topology, mantissa: int = 4) -> float:
        """
        Returns the distance between the input vertex and the input topology.

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex.
        topology : topologic.Topology
            The input topology.
        mantissa: int , optional
            The desired length of the mantissa. The default is 4.

        Returns
        -------
        float
            The distance between the input vertex and the input topology.

        """
        if not isinstance(vertex, topologic.Vertex) or not isinstance(topology, topologic.Topology):
            return None
        return round(topologic.VertexUtility.Distance(vertex, topology), mantissa)
    
    @staticmethod
    def EnclosingCell(vertex: topologic.Vertex, topology: topologic.Topology, exclusive: bool = True, tolerance: float = 0.0001) -> list:
        """
        Returns the list of Cells found in the input topology that enclose the input vertex.

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex.
        topology : topologic.Topology
            The input topology.
        exclusive : bool , optional
            If set to True, return only the first found enclosing cell. The default is True.
        tolerance : float , optional
            The tolerance for computing if the input vertex is enclosed in a cell. The default is 0.0001.

        Returns
        -------
        list
            The list of enclosing cells.

        """
        
        def boundingBox(cell):
            vertices = []
            _ = cell.Vertices(None, vertices)
            x = []
            y = []
            z = []
            for aVertex in vertices:
                x.append(aVertex.X())
                y.append(aVertex.Y())
                z.append(aVertex.Z())
            return ([min(x), min(y), min(z), max(x), max(y), max(z)])
        
        if isinstance(topology, topologic.Cell):
            cells = [topology]
        elif isinstance(topology, topologic.Cluster) or isinstance(topology, topologic.CellComplex):
            cells = []
            _ = topology.Cells(None, cells)
        else:
            return None
        if len(cells) < 1:
            return None
        enclosingCells = []
        for i in range(len(cells)):
            bbox = boundingBox(cells[i])
            if ((vertex.X() < bbox[0]) or (vertex.Y() < bbox[1]) or (vertex.Z() < bbox[2]) or (vertex.X() > bbox[3]) or (vertex.Y() > bbox[4]) or (vertex.Z() > bbox[5])) == False:
                if topologic.CellUtility.Contains(cells[i], vertex, tolerance) == 0:
                    if exclusive:
                        return([cells[i]])
                    else:
                        enclosingCells.append(cells[i])
        return enclosingCells

    @staticmethod
    def Index(vertex: topologic.Vertex, vertices: list, strict: bool = False, tolerance: float = 0.0001) -> int:
        """
        Returns index of the input vertex in the input list of vertices

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex.
        vertices : list
            The input list of vertices.
        strict : bool , optional
            If set to True, the vertex must be strictly identical to the one found in the list. Otherwise, a distance comparison is used. The default is False.
        tolerance : float , optional
            The tolerance for computing if the input vertex is enclosed in a cell. The default is 0.0001.

        Returns
        -------
        int
            The index of the input vertex in the input list of vertices.

        """
        from topologicpy.Topology import Topology
        if not isinstance(vertex, topologic.Vertex):
            return None
        if not isinstance(vertices, list):
            return None
        new_vertices = [v for v in vertices if isinstance(v, topologic.Vertex)]
        if not len(new_vertices) == len(vertices):
            return None
        for i in range(len(vertices)):
            if strict:
                if Topology.IsSame(vertex, vertices[i]):
                    return i
            else:
                d = Vertex.Distance(vertex, vertices[i])
                if d < tolerance:
                    return i
        return None

    @staticmethod
    def IsInside(vertex: topologic.Vertex, topology: topologic.Topology, tolerance: float = 0.0001) -> bool:
        """
        Returns True if the input vertex is inside the input topology. Returns False otherwise.

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex.
        topology : topologic.Topology
            The input topology.
        tolerance : float , optional
            The tolerance for computing if the input vertex is enclosed in a cell. The default is 0.0001.

        Returns
        -------
        bool
            True if the input vertex is inside the input topology. False otherwise.

        """
        from topologicpy.Wire import Wire
        from topologicpy.Shell import Shell
        from topologicpy.CellComplex import CellComplex
        from topologicpy.Cluster import Cluster
        from topologicpy.Topology import Topology
        if not isinstance(vertex, topologic.Vertex):
            return None
        if not isinstance(topology, topologic.Topology):
            return None

        if isinstance(topology, topologic.Vertex):
            return topologic.VertexUtility.Distance(vertex, topology) < tolerance
        elif isinstance(topology, topologic.Edge):
            try:
                parameter = topologic.EdgeUtility.ParameterAtPoint(topology, vertex)
            except:
                parameter = 400 #aribtrary large number greater than 1
            return 0 <= parameter <= 1
        elif isinstance(topology, topologic.Wire):
            edges = Wire.Edges(topology)
            for edge in edges:
                if Vertex.IsInside(vertex, edge, tolerance):
                    return True
            return False
        elif isinstance(topology, topologic.Face):
            return topologic.FaceUtility.IsInside(topology, vertex, tolerance)
        elif isinstance(topology, topologic.Shell):
            faces = Shell.Faces(topology)
            for face in faces:
                if Vertex.IsInside(vertex, face, tolerance):
                    return True
            return False
        elif isinstance(topology, topologic.Cell):
            return topologic.CellUtility.Contains(topology, vertex, tolerance) == 0
        elif isinstance(topology, topologic.CellComplex):
            cells = CellComplex.Cells(topology)
            faces = CellComplex.Faces(topology)
            edges = CellComplex.Edges(topology)
            vertices = CellComplex.Vertices(topology)
            subtopologies = cells + faces + edges + vertices
            for subtopology in subtopologies:
                if Vertex.IsInside(vertex, subtopology, tolerance):
                    return True
            return False
        elif isinstance(topology, topologic.Cluster):
            cells = Cluster.Cells(topology)
            faces = Cluster.Faces(topology)
            edges = Cluster.Edges(topology)
            vertices = Cluster.Vertices(topology)
            subtopologies = cells + faces + edges + vertices
            for subtopology in subtopologies:
                if Vertex.IsInside(vertex, subtopology, tolerance):
                    return True
            return False
        return False
    
    @staticmethod
    def NearestVertex(vertex: topologic.Vertex, topology: topologic.Topology, useKDTree: bool = True) -> topologic.Vertex:
        """
        Returns the vertex found in the input topology that is the nearest to the input vertex.

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex.
        topology : topologic.Topology
            The input topology to be searched for the nearest vertex.
        useKDTree : bool , optional
            if set to True, the algorithm will use a KDTree method to search for the nearest vertex. The default is True.

        Returns
        -------
        topologic.Vertex
            The nearest vertex.

        """        
        def SED(a, b):
            """Compute the squared Euclidean distance between X and Y."""
            p1 = (a.X(), a.Y(), a.Z())
            p2 = (b.X(), b.Y(), b.Z())
            return sum((i-j)**2 for i, j in zip(p1, p2))
        
        BT = collections.namedtuple("BT", ["value", "left", "right"])
        BT.__doc__ = """
        A Binary Tree (BT) with a node value, and left- and
        right-subtrees.
        """
        def firstItem(v):
            return v.X()
        def secondItem(v):
            return v.Y()
        def thirdItem(v):
            return v.Z()

        def itemAtIndex(v, index):
            if index == 0:
                return v.X()
            elif index == 1:
                return v.Y()
            elif index == 2:
                return v.Z()

        def sortList(vertices, index):
            if index == 0:
                vertices.sort(key=firstItem)
            elif index == 1:
                vertices.sort(key=secondItem)
            elif index == 2:
                vertices.sort(key=thirdItem)
            return vertices
        
        def kdtree(topology):
            assert isinstance(topology, topologic.Topology), "Vertex.NearestVertex: The input is not a Topology."
            vertices = []
            _ = topology.Vertices(None, vertices)
            assert (len(vertices) > 0), "Vertex.NearestVertex: Could not find any vertices in the input Topology"

            """Construct a k-d tree from an iterable of vertices.

            This algorithm is taken from Wikipedia. For more details,

            > https://en.wikipedia.org/wiki/K-d_tree#Construction

            """
            # k = len(points[0])
            k = 3

            def build(*, vertices, depth):
                if len(vertices) == 0:
                    return None
                #points.sort(key=operator.itemgetter(depth % k))
                vertices = sortList(vertices, (depth % k))

                middle = len(vertices) // 2
                
                return BT(
                    value = vertices[middle],
                    left = build(
                        vertices=vertices[:middle],
                        depth=depth+1,
                    ),
                    right = build(
                        vertices=vertices[middle+1:],
                        depth=depth+1,
                    ),
                )

            return build(vertices=list(vertices), depth=0)
        
        NNRecord = collections.namedtuple("NNRecord", ["vertex", "distance"])
        NNRecord.__doc__ = """
        Used to keep track of the current best guess during a nearest
        neighbor search.
        """

        def find_nearest_neighbor(*, tree, vertex):
            """Find the nearest neighbor in a k-d tree for a given vertex.
            """
            k = 3 # Forcing k to be 3 dimensional
            best = None
            def search(*, tree, depth):
                """Recursively search through the k-d tree to find the nearest neighbor.
                """
                nonlocal best

                if tree is None:
                    return
                distance = SED(tree.value, vertex)
                if best is None or distance < best.distance:
                    best = NNRecord(vertex=tree.value, distance=distance)

                axis = depth % k
                diff = itemAtIndex(vertex,axis) - itemAtIndex(tree.value,axis)
                if diff <= 0:
                    close, away = tree.left, tree.right
                else:
                    close, away = tree.right, tree.left

                search(tree=close, depth=depth+1)
                if diff**2 < best.distance:
                    search(tree=away, depth=depth+1)

            search(tree=tree, depth=0)
            return best.vertex
        
        if useKDTree:
            tree = kdtree(topology)
            return find_nearest_neighbor(tree=tree, vertex=vertex)
        else:
            vertices = []
            _ = topology.Vertices(None, vertices)
            distances = []
            indices = []
            for i in range(len(vertices)):
                distances.append(SED(vertex, vertices[i]))
                indices.append(i)
            sorted_indices = [x for _, x in sorted(zip(distances, indices))]
        return vertices[sorted_indices[0]]

    @staticmethod
    def Origin() -> topologic.Vertex:
        """
        Returns a vertex with coordinates (0,0,0)

        Parameters
        -----------

        Return
        -----------
        topologic.Vertex
        """
        return Vertex.ByCoordinates(0,0,0)
    
    @staticmethod
    def Project(vertex: topologic.Vertex, face: topologic.Face, direction: bool = None, mantissa: int = 4, tolerance: float = 0.0001) -> topologic.Vertex:
        """
        Returns a vertex that is the projection of the input vertex unto the input face.

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex to project unto the input face.
        face : topologic.Face
            The input face that receives the projection of the input vertex.
        direction : vector, optional
            The direction in which to project the input vertex unto the input face. If not specified, the direction of the projection is the normal of the input face. The default is None.
        mantissa : int , optional
            The length of the desired mantissa. The default is 4.
        tolerance : float , optional
            The desired tolerance. The default is 0.0001.

        Returns
        -------
        topologic.Vertex
            The projected vertex.

        """
        from topologicpy.Edge import Edge
        from topologicpy.Face import Face
        from topologicpy.Topology import Topology
        from topologicpy.Vector import Vector

        if not isinstance(vertex, topologic.Vertex):
            return None
        if not isinstance(face, topologic.Face):
            return None
        if not direction:
            direction = Vector.Reverse(Face.NormalAtParameters(face, 0.5, 0.5, "XYZ", mantissa))
        if topologic.FaceUtility.IsInside(face, vertex, tolerance):
            return vertex
        d = topologic.VertexUtility.Distance(vertex, face)*10
        far_vertex = topologic.TopologyUtility.Translate(vertex, direction[0]*d, direction[1]*d, direction[2]*d)
        if topologic.VertexUtility.Distance(vertex, far_vertex) > tolerance:
            e = topologic.Edge.ByStartVertexEndVertex(vertex, far_vertex)
            pv = face.Intersect(e, False)
            if not pv:
                far_vertex = topologic.TopologyUtility.Translate(vertex, -direction[0]*d, -direction[1]*d, -direction[2]*d)
                if topologic.VertexUtility.Distance(vertex, far_vertex) > tolerance:
                    e = topologic.Edge.ByStartVertexEndVertex(vertex, far_vertex)
                    pv = face.Intersect(e, False)
            return pv
        else:
            return None

    @staticmethod
    def X(vertex: topologic.Vertex, mantissa: int = 4) -> float:
        """
        Returns the X coordinate of the input vertex.

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex.
        mantissa : int , optional
            The desired length of the mantissa. The default is 4.

        Returns
        -------
        float
            The X coordinate of the input vertex.

        """
        if not isinstance(vertex, topologic.Vertex):
            return None
        return round(vertex.X(), mantissa)

    @staticmethod
    def Y(vertex: topologic.Vertex, mantissa: int = 4) -> float:
        """
        Returns the Y coordinate of the input vertex.

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex.
        mantissa : int , optional
            The desired length of the mantissa. The default is 4.

        Returns
        -------
        float
            The Y coordinate of the input vertex.

        """
        if not isinstance(vertex, topologic.Vertex):
            return None
        return round(vertex.Y(), mantissa)

    @staticmethod
    def Z(vertex: topologic.Vertex, mantissa: int = 4) -> float:
        """
        Returns the Z coordinate of the input vertex.

        Parameters
        ----------
        vertex : topologic.Vertex
            The input vertex.
        mantissa : int , optional
            The desired length of the mantissa. The default is 4.

        Returns
        -------
        float
            The Z coordinate of the input vertex.

        """
        if not isinstance(vertex, topologic.Vertex):
            return None
        return round(vertex.Z(), mantissa)
           


import topologicpy
import topologic


class Grid(topologic.Cluster):
    @staticmethod
    def EdgesByDistances(face=None, uOrigin=None, vOrigin=None, uRange=[-0.5,-0.25,0, 0.25,0.5], vRange=[-0.5,-0.25,0, 0.25,0.5], clip=False, tolerance=0.0001):
        """
        Creates a grid (cluster of edges).

        Parameters
        ----------
        face : topologic.Face , optional
            The input face. If set to None, the grid will be created on the XY plane. The default is None.
        uOrigin : topologic.Vertex , optional
            The origin of the *u* grid lines. If set to None: if the face is set, the uOrigin will be set to vertex at the face's 0,0 paratmer. If the face is set to None, the uOrigin will be set to the origin. The default is None.
        vOrigin : topologic.Vertex , optional
            The origin of the *v* grid lines. If set to None: if the face is set, the vOrigin will be set to vertex at the face's 0,0 paratmer. If the face is set to None, the vOrigin will be set to the origin. The default is None.
        uRange : list , optional
            A list of distances for the *u* grid lines from the uOrigin. The default is [-0.5,-0.25,0, 0.25,0.5].
        vRange : list , optional
            A list of distances for the *v* grid lines from the vOrigin. The default is [-0.5,-0.25,0, 0.25,0.5].
        clip : bool , optional
            If True the grid will be clipped by the shape of the input face. The default is False.
        tolerance : float , optional
            The desired tolerance. The default is 0.0001.

        Returns
        -------
        topologic.Cluster
            The created grid. Edges in the grid have an identifying dictionary with two keys: "dir" and "offset". The "dir" key can have one of two values: "u" or "v", the "offset" key contains the offset distance of that grid edge from the specified origin.

        """
        from topologicpy.Vertex import Vertex
        from topologicpy.Edge import Edge
        from topologicpy.Face import Face
        from topologicpy.Cluster import Cluster
        from topologicpy.Topology import Topology
        from topologicpy.Dictionary import Dictionary
        from topologicpy.Vector import Vector
        if len(uRange) < 1 or len(vRange) < 1:
            return None
        if not uOrigin:
            if not isinstance(face, topologic.Face):
                uOrigin = Vertex.ByCoordinates(0,0,0)
            else:
                uOrigin = Face.VertexByParameters(face, 0, 0)
        if not vOrigin:
            if not isinstance(face, topologic.Face):
                vOrigin = Vertex.ByCoordinates(0,0,0)
            else:
                vOrigin = Face.VertexByParameters(face, 0, 0)
        
        if isinstance(face, topologic.Face):
            v1 = Face.VertexByParameters(face, 0, 0)
            v2 = Face.VertexByParameters(face, 1, 0)
            v3 = Face.VertexByParameters(face, 0, 0)
            v4 = Face.VertexByParameters(face, 0, 1)
        else:
            v1 = Vertex.ByCoordinates(0,0,0)
            v2 = Vertex.ByCoordinates(max(uRange),0,0)
            v3 = Vertex.ByCoordinates(0,0,0)
            v4 = Vertex.ByCoordinates(0,max(vRange),0)

        uVector = [v2.X()-v1.X(), v2.Y()-v1.Y(),v2.Z()-v1.Z()]
        vVector = [v4.X()-v3.X(), v4.Y()-v3.Y(),v4.Z()-v3.Z()]
        gridEdges = []
        if len(uRange) > 0:
            uRange.sort()
            uuVector = Vector.Normalize(uVector)
            for u in uRange:
                tempVec = Vector.Multiply(uuVector, u, tolerance)
                v1 = Vertex.ByCoordinates(uOrigin.X()+tempVec[0], uOrigin.Y()+tempVec[1], uOrigin.Z()+tempVec[2])
                v2 = Vertex.ByCoordinates(v1.X()+vVector[0], v1.Y()+vVector[1], v1.Z()+vVector[2])
                e = Edge.ByVertices([v1, v2])
                if clip and isinstance(face, topologic.Face):
                    e = e.Intersect(face, False)
                if e:
                    if isinstance(e, topologic.Edge):
                        d = Dictionary.ByKeysValues(["dir", "offset"],["u",u])
                        e.SetDictionary(d)
                        gridEdges.append(e)
                    elif e.Type() > topologic.Edge.Type():
                        tempEdges = []
                        _ = e.Edges(None, tempEdges)
                        for tempEdge in tempEdges:
                            d = Dictionary.ByKeysValues(["dir", "offset"],["u",u])
                            tempEdge.SetDictionary(d)
                            gridEdges.append(tempEdge)
        if len(vRange) > 0:
            vRange.sort()
            uvVector = Vector.Normalize(vVector)
            for v in vRange:
                tempVec = Vector.Multiply(uvVector, v, tolerance)
                v1 = Vertex.ByCoordinates(vOrigin.X()+tempVec[0], vOrigin.Y()+tempVec[1], vOrigin.Z()+tempVec[2])
                v2 = Vertex.ByCoordinates(v1.X()+uVector[0], v1.Y()+uVector[1], v1.Z()+uVector[2])
                e = Edge.ByVertices([v1, v2])
                if clip and isinstance(face, topologic.Face):
                    e = e.Intersect(face, False)
                if e:
                    if isinstance(e, topologic.Edge):
                        d = Dictionary.ByKeysValues(["dir", "offset"],["v",v])
                        e.SetDictionary(d)
                        gridEdges.append(e)
                    elif e.Type() > topologic.Edge.Type():
                        tempEdges = []
                        _ = e.Edges(None, tempEdges)
                        for tempEdge in tempEdges:
                            d = Dictionary.ByKeysValues(["dir", "offset"],["v",v])
                            tempEdge.SetDictionary(d)
                            gridEdges.append(tempEdge)
        grid = None
        if len(gridEdges) > 0:
            grid = Cluster.ByTopologies(gridEdges)
        return grid
    
    @staticmethod
    def EdgesByParameters(face, uRange=[0,0.25,0.5,0.75,1.0], vRange=[0,0.25,0.5,0.75,1.0], clip=False):
        """
        Creates a grid (cluster of edges).

        Parameters
        ----------
        face : topologic.Face
            The input face.
        uRange : list , optional
            A list of *u* parameters for the *u* grid lines. The default is [0,0.25,0.5, 0.75, 1.0].
        vRange : list , optional
            A list of *v* parameters for the *v* grid lines. The default is [0,0.25,0.5, 0.75, 1.0].
        clip : bool , optional
            If True the grid will be clipped by the shape of the input face. The default is False.

        Returns
        -------
        topologic.Cluster
            The created grid. Edges in the grid have an identifying dictionary with two keys: "dir" and "offset". The "dir" key can have one of two values: "u" or "v", the "offset" key contains the offset parameter of that grid edge.

        """
        from topologicpy.Vertex import Vertex
        from topologicpy.Edge import Edge
        from topologicpy.Face import Face
        from topologicpy.Cluster import Cluster
        from topologicpy.Dictionary import Dictionary

        if not isinstance(face, topologic.Face):
            return None
        if len(uRange) < 1 and len(vRange) < 1:
            return None
        if len(uRange) > 0:
            if (min(uRange) < 0) or (max(uRange) > 1):
                return None
        if len(vRange) > 0:
            if (min(vRange) < 0) or (max(vRange) > 1):
                return None

        uRange.sort()
        vRange.sort()
        gridEdges = []
        for u in uRange:
            v1 = Face.VertexByParameters(face, u, 0)
            v2 = Face.VertexByParameters(face, u, 1)
            e = Edge.ByVertices([v1, v2])
            if clip and isinstance(face, topologic.Face):
                e = e.Intersect(face, False)
            if e:
                if isinstance(e, topologic.Edge):
                    d = Dictionary.ByKeysValues(["dir", "offset"],["u",u])
                    e.SetDictionary(d)
                    gridEdges.append(e)
                elif e.Type() > topologic.Edge.Type():
                    tempEdges = []
                    _ = e.Edges(None, tempEdges)
                    for tempEdge in tempEdges:
                        d = Dictionary.ByKeysValues(["dir", "offset"],["u",u])
                        tempEdge.SetDictionary(d)
                        gridEdges.append(tempEdge)
        for v in vRange:
            v1 = Face.VertexByParameters(face, 0, v)
            v2 = Face.VertexByParameters(face, 1, v)
            e = Edge.ByVertices([v1, v2])
            if clip and isinstance(face, topologic.Face):
                e = e.Intersect(face, False)
            if e:
                if isinstance(e, topologic.Edge):
                    d = Dictionary.ByKeysValues(["dir", "offset"],["v",v])
                    e.SetDictionary(d)
                    gridEdges.append(e)
                elif e.Type() > topologic.Edge.Type():
                    tempEdges = []
                    _ = e.Edges(None, tempEdges)
                    for tempEdge in tempEdges:
                        d = Dictionary.ByKeysValues(["dir", "offset"],["v",v])
                        tempEdge.SetDictionary(d)
                        gridEdges.append(tempEdge)
        grid = None
        if len(gridEdges) > 0:
            grid = Cluster.ByTopologies(gridEdges)
        return grid


    @staticmethod
    def VerticesByDistances(face=None, origin=None, uRange=[-0.5,-0.25,0, 0.25,0.5], vRange=[-0.5,-0.25,0,0.25,0.5], clip=False, tolerance=0.0001):
        """
        Creates a grid (cluster of vertices).

        Parameters
        ----------
        face : topologic.Face , optional
            The input face. If set to None, the grid will be created on the XY plane. The default is None.
        origin : topologic.Vertex , optional
            The origin of the grid vertices. If set to None: if the face is set, the origin will be set to vertex at the face's 0,0 paratmer. If the face is set to None, the origin will be set to (0,0,0). The default is None.
        uRange : list , optional
            A list of distances for the *u* grid lines from the uOrigin. The default is [-0.5,-0.25,0, 0.25,0.5].
        vRange : list , optional
            A list of distances for the *v* grid lines from the vOrigin. The default is [-0.5,-0.25,0, 0.25,0.5].
        clip : bool , optional
            If True the grid will be clipped by the shape of the input face. The default is False.
        tolerance : float , optional
            The desired tolerance. The default is 0.0001.

        Returns
        -------
        topologic.Cluster
            The created grid. Vertices in the grid have an identifying dictionary with two keys: "u" and "v". The "dir" key can have one of two values: "u" or "v" that contain the *u* and *v* offset distances of that grid vertex from the specified origin.

        """
        from topologicpy.Vertex import Vertex
        from topologicpy.Edge import Edge
        from topologicpy.Face import Face
        from topologicpy.Cluster import Cluster
        from topologicpy.Topology import Topology
        from topologicpy.Dictionary import Dictionary
        from topologicpy.Vector import Vector
        if len(uRange) < 1 or len(vRange) < 1:
            return None
        if not origin:
            if not isinstance(face, topologic.Face):
                origin = Vertex.ByCoordinates(0,0,0)
            else:
                origin = Face.VertexByParameters(face, 0, 0)
        
        if isinstance(face, topologic.Face):
            v1 = Face.VertexByParameters(face, 0, 0)
            v2 = Face.VertexByParameters(face, 1, 0)
            v3 = Face.VertexByParameters(face, 0, 0)
            v4 = Face.VertexByParameters(face, 0, 1)
        else:
            v1 = Vertex.ByCoordinates(0,0,0)
            v2 = Vertex.ByCoordinates(max(uRange),0,0)
            v3 = Vertex.ByCoordinates(0,0,0)
            v4 = Vertex.ByCoordinates(0,max(vRange),0)

        uVector = [v2.X()-v1.X(), v2.Y()-v1.Y(),v2.Z()-v1.Z()]
        vVector = [v4.X()-v3.X(), v4.Y()-v3.Y(),v4.Z()-v3.Z()]
        gridVertices = []
        if len(uRange) > 0:
            uRange.sort()
            uuVector = Vector.Normalize(uVector)
            uvVector = Vector.Normalize(vVector)
            for u in uRange:
                for v in vRange:
                    uTempVec = Vector.Multiply(uuVector, u, tolerance)
                    vTempVec = Vector.Multiply(uvVector, v, tolerance)
                    gridVertex = Vertex.ByCoordinates(origin.X()+uTempVec[0], origin.Y()+vTempVec[1], origin.Z()+uTempVec[2])
                    if clip and isinstance(face, topologic.Face):
                        gridVertex = gridVertex.Intersect(face, False)
                    if isinstance(gridVertex, topologic.Vertex):
                        d = Dictionary.ByKeysValues(["u","v"],[u,v])
                        if d:
                            gridVertex.SetDictionary(d)
                        gridVertices.append(gridVertex)
        grid = None
        if len(gridVertices) > 0:
            grid = Cluster.ByTopologies(gridVertices)
        return grid

    @staticmethod
    def VerticesByParameters(face=None, uRange=[0.0,0.25,0.5,0.75,1.0], vRange=[0.0,0.25,0.5,0.75,1.0], clip=False, tolerance=0.0001):
        """
        Creates a grid (cluster of vertices).

        Parameters
        ----------
        face : topologic.Face , optional
            The input face. If set to None, the grid will be created on the XY plane. The default is None.
        origin : topologic.Vertex , optional
            The origin of the grid vertices. If set to None: if the face is set, the origin will be set to vertex at the face's 0,0 paratmer. If the face is set to None, the origin will be set to (0,0,0). The default is None.
        uRange : list , optional
            A list of *u* parameters for the *u* grid lines from the uOrigin. The default is [0.0,0.25,0.5,0.75,1.0].
        vRange : list , optional
            A list of *v* parameters for the *v* grid lines from the vOrigin. The default is [0.0,0.25,0.5,0.75,1.0].
        clip : bool , optional
            If True the grid will be clipped by the shape of the input face. The default is False.
        tolerance : float , optional
            The desired tolerance. The default is 0.0001.

        Returns
        -------
        topologic.Cluster
            The created grid. Vertices in the grid have an identifying dictionary with two keys: "u" and "v". The "dir" key can have one of two values: "u" or "v" that contain the *u* and *v* offset distances of that grid vertex from the specified origin.

        """
        from topologicpy.Vertex import Vertex
        from topologicpy.Edge import Edge
        from topologicpy.Face import Face
        from topologicpy.Cluster import Cluster
        from topologicpy.Topology import Topology
        from topologicpy.Dictionary import Dictionary
        from topologicpy.Vector import Vector

        if not isinstance(face, topologic.Face):
            return None
        if len(uRange) < 1 or len(vRange) < 1:
            return None
        if (min(uRange) < 0) or (max(uRange) > 1):
            return None
        if (min(vRange) < 0) or (max(vRange) > 1):
            return None

        uRange.sort()
        vRange.sort()
        gridVertices = []
        if len(uRange) > 0:
            uRange.sort()
            for u in uRange:
                for v in vRange:
                    gridVertex = Vertex.ByParameter(face, u, v)
                    if clip and isinstance(face, topologic.Face):
                        gridVertex = gridVertex.Intersect(face, False)
                    if isinstance(gridVertex, topologic.Vertex):
                        d = Dictionary.ByKeysValues(["u","v"],[u,v])
                        if d:
                            gridVertex.SetDictionary(d)
                        gridVertices.append(gridVertex)
        grid = None
        if len(gridVertices) > 0:
            grid = Cluster.ByTopologies(gridVertices)
        return grid